
/******************************************************************************
* MODULE     : linear_anfunc.hpp
* DESCRIPTION: Vectors of analytic functions
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_LINEAR_ANFUNC_HPP
#define __MML_LINEAR_ANFUNC_HPP
#include "anfunc.hpp"
#include "linear_series.hpp"
namespace mml {
#define TMPL_DEF template<typename C, typename V, typename W>
#define TMPL template<typename C, typename V, typename W>
#define Anfunc anfunc<C,V>
#define Anfunc_rep anfunc_rep<C,V>
#define Assume assume_bound<typename unvectorize<Error >::val>
#define Scalar typename unvectorize<C>::val
#define Series_variant naive_multiply
#define Series series<C,Series_variant>
#define Error Abs_type(C)

/******************************************************************************
* Prototypes
******************************************************************************/

TMPL Anfunc access (const anfunc<vector<C,W>,V>& f, nat i);
TMPL vector<Anfunc,W> as_vector (const anfunc<vector<C,W>,V>& f);
TMPL anfunc<vector<C,W>,V> as_anfunc (const vector<Anfunc,W>& v);

TMPL anfunc<vector<C,W>,V> trig (const anfunc<vector<C,W>,V>& f);
TMPL vector<Anfunc> cos_sin (const Anfunc& f);
template<typename C,typename V> Anfunc cos (const Anfunc& f);
template<typename C,typename V> Anfunc sin (const Anfunc& f);
template<typename C,typename V> Anfunc tan (const Anfunc& f);

TMPL anfunc<vector<C,W>,V>
solve_vector_lde_init (const anfunc<vector<C,W>,V>& f, const vector<C,W>& c);
TMPL vector<Anfunc,W>
solve_lde_init (const vector<Anfunc,W>& f, const vector<C,W>& c);

TMPL Anfunc access (const anfunc<matrix<C,W>,V>& f, nat i, nat j);
TMPL matrix<Anfunc,W> as_matrix (const anfunc<matrix<C,W>,V>& f);
TMPL anfunc<matrix<C,W>,V> as_anfunc (const matrix<Anfunc,W>& m);

TMPL anfunc<matrix<C,W>,V>
solve_matrix_lde_init (const anfunc<matrix<C,W>,V>& f, const matrix<C,W>& c);
TMPL matrix<Anfunc,W>
solve_lde (const matrix<Anfunc,W>& f);
TMPL matrix<Anfunc,W>
solve_lde_init (const matrix<Anfunc,W>& f, const matrix<C,W>& c);

/******************************************************************************
* Vectors of analytic functions
******************************************************************************/

#define Anfunc_vector anfunc<vector<C,W>,V>
#define Vector_anfunc vector<anfunc<C,V>,W>
STYPE_TO_TYPE(TMPL,as_vector_type,Anfunc_vector,Vector_anfunc);
#undef Vector_anfunc
#undef Anfunc_vector

template<typename C, typename W>
struct unvectorize<vector<C,W> > {
  // FIXME: recursion, etc.
  typedef C val;
  static inline vector<val> encode (const vector<C,W>& x) {
    return x; }
  static inline vector<C,W>
  decode (const vector<val>& v, const vector<C,W>& gauge) {
    (void) gauge;
    return v; }
};

TMPL
class vector_access_anfunc_rep : public Anfunc_rep {
protected:
  anfunc<vector<C,W>,V> f;
  nat i;
public:
  vector_access_anfunc_rep (const anfunc<vector<C,W>,V>& f2, nat i2):
    f (f2), i (i2) {}
  Series Expand () const {
    return access (expand (f), i); }
  Error Tail_bound (nat order, Assume& a) const {
    return tail_bound (this->f, order, a) [i]; }
  Anfunc Focus (const Scalar& z) const {
    return access (focus (f, z), i); }
  Anfunc Move (const Scalar& z) const {
    return access (move (f, z), i); }
  Anfunc Derive () const {
    return access (derive (f), i); }
};

TMPL Anfunc
access (const anfunc<vector<C,W>,V>& f, nat i) {
  return (Anfunc_rep*) new vector_access_anfunc_rep<C,V,W> (f, i);
}

TMPL vector<Anfunc,W>
as_vector (const anfunc<vector<C,W>,V>& f) {
  nat n= dim (f[0]);
  vector<Anfunc,W> v (Anfunc (0), n);
  for (nat i=0; i<n; i++)
    v (i)= access (f, i);
  return v;
}

TMPL
class vector_anfunc_rep : public anfunc_rep<vector<C,W>,V> {
  vector<Anfunc,W> v;
public:
  vector_anfunc_rep (const vector<Anfunc,W>& v2): v (v2) {}
  series<vector<C,W>,Series_variant> Expand () const {
    nat i, n= dim (v);
    vector<Series,W> r (Series (0), n);
    for (i=0; i<n; i++)
      r(i)= expand (v[i]);
    return as_series (r); }
  vector<Error,W> Tail_bound (nat order, Assume& a) const {
    nat i, n= dim (v);
    vector<Error,W> r (Error (0), n);
    for (i=0; i<n; i++)
      r(i)= tail_bound (v[i], order, a);
    return r; }
  anfunc<vector<C,W>,V> Focus (const Scalar& z) const {
    nat i, n= dim (v);
    vector<Anfunc,W> r (Anfunc (0), n);
    for (i=0; i<n; i++)
      r(i)= focus (v[i], z);
    return as_anfunc (r); }
  anfunc<vector<C,W>,V> Move (const Scalar& z) const {
    nat i, n= dim (v);
    vector<Anfunc,W> r (Anfunc (0), n);
    for (i=0; i<n; i++)
      r(i)= move (v[i], z);
    return as_anfunc (r); }
  anfunc<vector<C,W>,V> Derive () const {
    return as_anfunc (derive (v)); }
};

TMPL anfunc<vector<C,W>,V>
as_anfunc (const vector<Anfunc,W>& v) {
  return (anfunc_rep<vector<C,W>,V>*) new vector_anfunc_rep<C,V,W> (v);
}

TMPL anfunc<vector<C,W>,V>
from_vector (const vector<Anfunc,W>& v) {
  return (anfunc_rep<vector<C,W>,V>*) new vector_anfunc_rep<C,V,W> (v);
}

/******************************************************************************
* Trigonometric functions and linear differential equations
******************************************************************************/

TMPL anfunc<vector<C,W>,V>
trig (const anfunc<vector<C,W>,V>& f) {
  return unary_implicit_anfunc<trig_op> (f);
}

template<typename C,typename V> vector<Anfunc >
cos_sin (const Anfunc& f) {
  return as_vector (trig (as_anfunc (vec (f))));
}

template<typename C,typename V> Anfunc
cos (const Anfunc& f) {
  return unary_implicit_anfunc<cos_op> (f);
}

template<typename C,typename V> Anfunc
sin (const Anfunc& f) {
  return unary_implicit_anfunc<sin_op> (f);
}

template<typename C,typename V> Anfunc
tan (const Anfunc& f) {
  return unary_implicit_anfunc<tan_op> (f);
}

TMPL anfunc<vector<C,W>,V>
solve_vector_lde_init (const anfunc<vector<C,W>,V>& f, const vector<C,W>& c) {
  return unary_implicit_anfunc<solve_vector_lde_op> (f, c);
}

TMPL vector<Anfunc,W>
solve_lde_init (const vector<Anfunc,W>& f, const vector<C,W>& c) {
  return as_vector (solve_vector_lde_init (as_anfunc (f), c));
}

/******************************************************************************
* Matrices of analytic functions
******************************************************************************/

#define Anfunc_matrix anfunc<matrix<C,W>,V>
#define Matrix_anfunc matrix<anfunc<C,V>,W>
STYPE_TO_TYPE(TMPL,as_matrix_type,Anfunc_matrix,Matrix_anfunc);
#undef Matrix_anfunc
#undef Anfunc_matrix

template<typename C, typename W>
struct unvectorize<matrix<C,W> > {
  // FIXME: recursion, etc.
  typedef C val;
  static inline vector<val> encode (const matrix<C,W>& x) {
    nat nr= rows (x), nc= cols (x);
    vector<val> r (C(0), nr*nc);
    for (nat i=0; i<nr; i++)
      for (nat j=0; j<nc; j++)
	r(i*nc+j)= x(i,j);
    return r; }
  static inline matrix<C,W>
  decode (const vector<val>& v, const matrix<C,W>& gauge) {
    nat nr= rows (gauge), nc= cols (gauge);
    matrix<C,W> r (C(0), nr, nc);
    for (nat i=0; i<nr; i++)
      for (nat j=0; j<nc; j++)
	r.ref (i, j)= v[i*nc+j];
    return r; }
};

TMPL
class matrix_access_anfunc_rep : public Anfunc_rep {
protected:
  anfunc<matrix<C,W>,V> f;
  nat i, j;
public:
  matrix_access_anfunc_rep (const anfunc<matrix<C,W>,V>& f2, nat i2, nat j2):
    f (f2), i (i2), j (j2) {}
  Series Expand () const {
    return access (expand (f), i, j); }
  Error Tail_bound (nat order, Assume& a) const {
    return tail_bound (this->f, order, a) (i, j); }
  Anfunc Focus (const Scalar& z) const {
    return access (focus (f, z), i, j); }
  Anfunc Move (const Scalar& z) const {
    return access (move (f, z), i, j); }
  Anfunc Derive () const {
    return access (derive (f), i, j); }
};

TMPL Anfunc
access (const anfunc<matrix<C,W>,V>& f, nat i, nat j) {
  return (Anfunc_rep*) new matrix_access_anfunc_rep<C,V,W> (f, i, j);
}

TMPL matrix<Anfunc,W>
as_matrix (const anfunc<matrix<C,W>,V>& f) {
  nat nr= rows (f[0]), nc= cols (f[0]);
  matrix<Anfunc,W> m (Anfunc (0), nr, nc);
  for (nat i=0; i<nr; i++)
    for (nat j=0; j<nc; j++)
      m.ref (i, j)= access (f, i, j);
  return m;
}

TMPL
class matrix_anfunc_rep : public anfunc_rep<matrix<C,W>,V> {
  matrix<Anfunc,W> m;
public:
  matrix_anfunc_rep (const matrix<Anfunc,W>& m2): m (m2) {}
  series<matrix<C,W>,Series_variant> Expand () const {
    nat i, j, nr= rows (m), nc= cols (m);
    matrix<Series,W> r (Series (0), nr, nc);
    for (i=0; i<nr; i++)
      for (j=0; j<nc; j++)
	r.ref (i, j)= expand (m (i, j));
    return as_series (r); }
  matrix<Error,W> Tail_bound (nat order, Assume& a) const {
    nat i, j, nr= rows (m), nc= cols (m);
    matrix<Error,W> r (Error (0), nr, nc);
    for (i=0; i<nr; i++)
      for (j=0; j<nc; j++)
	r.ref (i, j)= tail_bound (m (i, j), order, a);
    return r; }
  anfunc<matrix<C,W>,V> Focus (const Scalar& z) const {
    nat i, j, nr= rows (m), nc= cols (m);
    matrix<Anfunc,W> r (Anfunc (0), nr, nc);
    for (i=0; i<nr; i++)
      for (j=0; j<nc; j++)
	r.ref (i, j)= focus (m (i, j), z);
    return as_anfunc (r); }
  anfunc<matrix<C,W>,V> Move (const Scalar& z) const {
    nat i, j, nr= rows (m), nc= cols (m);
    matrix<Anfunc,W> r (Anfunc (0), nr, nc);
    for (i=0; i<nr; i++)
      for (j=0; j<nc; j++)
	r.ref (i,j)= move (m (i, j), z);
    return as_anfunc (r); }
  anfunc<matrix<C,W>,V> Derive () const {
    return as_anfunc (derive (m)); }
};

TMPL anfunc<matrix<C,W>,V>
as_anfunc (const matrix<Anfunc,W>& m) {
  return (anfunc_rep<matrix<C,W>,V>*) new matrix_anfunc_rep<C,V,W> (m);
}

TMPL anfunc<matrix<C,W>,V>
from_matrix (const matrix<Anfunc,W>& m) {
  return (anfunc_rep<matrix<C,W>,V>*) new matrix_anfunc_rep<C,V,W> (m);
}

/******************************************************************************
* Linear differential matrix equations
******************************************************************************/

TMPL anfunc<matrix<C,W>,V>
solve_matrix_lde_init (const anfunc<matrix<C,W>,V>& f, const matrix<C,W>& c) {
  return unary_implicit_anfunc<solve_matrix_lde_op> (f, c);
}

TMPL matrix<Anfunc,W>
solve_lde (const matrix<Anfunc,W>& f) {
  assert (is_square_matrix (f));
  matrix<C,W> c (C(1), rows (f), cols (f));
  return as_matrix (solve_matrix_lde_init (as_anfunc (f), c));
}

TMPL matrix<Anfunc,W>
solve_lde_init (const matrix<Anfunc,W>& f, const matrix<C,W>& c) {
  assert (is_square_matrix (f) && rows (f) == rows (c));
  return as_matrix (solve_matrix_lde_init (as_anfunc (f), c));
}

#undef TMPL_DEF
#undef TMPL
#undef Anfunc
#undef Anfunc_rep
#undef Assume
#undef Scalar
#undef Series_variant
#undef Series
#undef Error
} // namespace mml
#endif // __MML_LINEAR_ANFUNC_HPP
