
/******************************************************************************
* MODULE     : mmlout.hpp
* DESCRIPTION: Pretty printing of expressions
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_MMLOUT_HPP
#define __MML_MMLOUT_HPP
namespace mml {

/******************************************************************************
* Printing expressions
******************************************************************************/

extern void* as_generic_aux_ptr;
extern string as_string (void* ptr);

template<typename C> string print_to_string (const C& x);
template<typename C> string print_to_lisp (const C& x);

template<typename C> generic
as_generic (const C& x) {
  if (as_generic_aux_ptr == (void*) &x)
    return print_to_string (x);
  else {
    void* old= as_generic_aux_ptr;
    as_generic_aux_ptr= (void*) &x;
    generic r= from_lisp (print_to_lisp (x));
    as_generic_aux_ptr= old;
    return r;
  }
}

template<typename C> inline generic
as_generic (C* const & x) {
  return as_string ((void*) x);
}

generic as_generic (int i);
generic as_generic (nat i);
generic as_generic (unsigned long int i);
generic as_generic (signed long int i);
generic as_generic (float x);
generic as_generic (double x);
generic as_generic (char* s);
generic as_generic (const string& s);
generic as_generic (const integer& i);
generic as_generic (const rational& r);
generic as_generic (const generic& r);

template<typename T> inline generic
as_generic (char* op, const T& t) {
  return generic (op) (as_generic (t));
}

template<typename T, typename U> inline generic
as_generic (char* op, const T& t, const U& u) {
  return generic (op) (as_generic (t), as_generic (u));
}

template<typename C> inline ostream&
print_lisp (ostream& out, const C& x) {
  lisp_print (out, as_generic (x));
  return out;
}

#define PRINT_SUGAR(TMPL,T)				\
TMPL inline ostream& operator << (ostream& out, const T& x) {	\
  return out << as_generic (x); }

} // namespace mml
#endif // __MML_MMLOUT_HPP
