
/******************************************************************************
* MODULE     : complex_ball.hpp
* DESCRIPTION: Standard complex balls
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MML_COMPLEX_BALL_HPP
#define __MML_COMPLEX_BALL_HPP
#include "real_ball.hpp"
#include "complexify.hpp"
#include "type_props.hpp"
namespace mml {
typedef complexify<real_ball_center> complex_ball_center;
typedef real_ball_radius complex_ball_radius;

class complex_ball {
MML_ALLOCATORS;
private:
  // Define fields of same size as "Ball" type (see implementation class)
  // IMPORTANT: The sizes MUST match for binary compatibility
  void* x;
  void* y;
  void* r;

public:
  complex_ball (const complex_ball& x);
  ~complex_ball ();
  complex_ball& operator = (const complex_ball& x);
  complex_ball ();
  complex_ball (int i);
  complex_ball (nat i);
  complex_ball (double x);
  complex_ball (const integer& i);
  complex_ball (const rational& x);
  complex_ball (const real_ball_center& x);
  complex_ball (const real_ball& x, const real_ball& y= 0);
  complex_ball (const complex_ball_center& z);
  complex_ball (const complex_ball_center& c, const complex_ball_radius& r);
};

STYPE_TO_TYPE(STMPL,real_type,complex_ball,real_ball);
STYPE_TO_TYPE(STMPL,abs_type,complex_ball,real_ball);
STYPE_TO_TYPE(STMPL,complex_type,real_ball,complex_ball);
STYPE_TO_TYPE(STMPL,center_type,complex_ball,complex_ball_center);
STYPE_TO_TYPE(STMPL,radius_type,complex_ball,complex_ball_radius);

STMPL
struct constant_helper<complex_ball > {
  static complex_ball nan ();
  static complex_ball plus_infinity ();
  static complex_ball minus_infinity ();
  static complex_ball infinity ();
  static complex_ball fuzz ();
  static complex_ball log2 ();
  static complex_ball pi ();
  static complex_ball euler ();
  static complex_ball imaginary ();
};

STMPL
struct make_ball_helper<complex_ball,complex_ball_center,complex_ball_radius> {
  static inline complex_ball
  val (const complex_ball_center& c, const complex_ball_radius& r) {
    return complex_ball (c, r); }
};

WRAP_BASIC_PROTO(,complex_ball);
WRAP_FIELD_PROTO(,complex_ball);
WRAP_SHIFT_PROTO(,complex_ball);
WRAP_RING_EQUAL_PROTO(,complex_ball);
WRAP_NUMBER_CHECK_PROTO(,complex_ball);
WRAP_SPECIAL_PROTO(,complex_ball);
WRAP_COMPLEX_PROTO(,complex_ball,real_ball);

generic as_generic (const complex_ball& x, nat dd);
complex_ball_center center (const complex_ball& z);
complex_ball_radius radius (const complex_ball& z);
complex_ball_radius abs_D (const complex_ball& x);
complex_ball_radius abs_U (const complex_ball& x);
template<> inline int var_sign (const complex_ball& z) {
  if (z == 0) return 0; else return 1; }

} // namespace mml
#endif // __MML_COMPLEX_BALL_HPP
