
/******************************************************************************
* MODULE     : symbolic.hpp
* DESCRIPTION: Symbolic expressions
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __SYMBOLIC_HPP
#define __SYMBOLIC_HPP
#include "universe.hpp"
#include "generic_internal.hpp"
namespace mml {
#define TMPL_DEF template<typename C,typename V=rules<exp_universe> >
#define TMPL template<typename C,typename V>
#define Symbolic symbolic<C,V>
#define Vector vector<Symbolic >
#define Table table<Symbolic,Symbolic >
TMPL class symbolic_rep;
TMPL class symbolic;

/******************************************************************************
* Symbolic class
******************************************************************************/

TMPL_DEF
class symbolic {
private:
  generic rep;

public:
  inline symbolic (const generic& g): rep (g) {}
  inline generic operator * () const { return rep; }

  inline symbolic ();
  inline symbolic (int i);
  inline symbolic (const C& sc);
  inline symbolic (char* s);
  inline symbolic (const string& s);
  inline symbolic (const Vector& v);

  inline Symbolic operator () () const;
  inline Symbolic operator () (const Symbolic& x1) const;
  inline Symbolic operator () (const Symbolic& x1, const Symbolic& x2) const;
  inline Symbolic operator () (const Symbolic& x1, const Symbolic& x2,
			       const Symbolic& x3) const;
  inline Symbolic operator [] (nat i) const { return rep[i]; }
};

/******************************************************************************
* Constructors
******************************************************************************/

TMPL inline generic constructor (const C& sc) {
  return generic_scalar<C> (sc); }
TMPL inline generic constructor (const string& s) {
  return generic_atom (s); }
TMPL inline generic constructor (const Vector& v) {
  return generic_compound (*reinterpret_cast<const vector<generic>*> (&v)); }

TMPL inline Symbolic::symbolic ():
  rep (constructor<C,V> (C(0))) {}
TMPL inline Symbolic::symbolic (int i):
  rep (constructor<C,V> (C(i))) {}
TMPL inline Symbolic::symbolic (const C& sc):
  rep (constructor<C,V> (sc)) {}
TMPL inline Symbolic::symbolic (char* s):
  rep (constructor<C,V> (string (s))) {}
TMPL inline Symbolic::symbolic (const string& s):
  rep (constructor<C,V> (s)) {}
TMPL inline Symbolic::symbolic (const Vector& v):
  rep (constructor<C,V> (v)) {}

TMPL inline Symbolic Symbolic::operator () () const {
  return constructor (vec (*this)); }
TMPL inline Symbolic Symbolic::operator () (const Symbolic& x1) const {
  return constructor (vec (*this, x1)); }
TMPL inline Symbolic Symbolic::operator ()
  (const Symbolic& x1, const Symbolic& x2) const {
    return constructor (vec (*this, x1, x2)); }
TMPL inline Symbolic Symbolic::operator ()
  (const Symbolic& x1, const Symbolic& x2, const Symbolic& x3) const {
    return constructor (vec (*this, x1, x2, x3)); }

TMPL inline Symbolic symbolic_sum (const Table& t) {
  return generic_sum (
    *reinterpret_cast<const table<generic,generic>*> (&t)); }
TMPL inline Symbolic symbolic_product (const Table& t) {
  return generic_product (
    *reinterpret_cast<const table<generic,generic>*> (&t)); }

/******************************************************************************
* Predicates and inspection
******************************************************************************/

TMPL inline bool operator == (const Symbolic& x1, const Symbolic& x2) {
  return (*x1) == (*x2); }
TMPL inline bool operator != (const Symbolic& x1, const Symbolic& x2) {
  return (*x1) != (*x2); }
TMPL inline bool operator == (const Symbolic& x, int i) {
  return (*x) == constructor<C,V> (C(i)); }
TMPL inline bool operator != (const Symbolic& x, int i) {
  return (*x) != constructor<C,V> (C(i)); }

TMPL inline bool is_scalar (const Symbolic& x) { return is_scalar (*x); }
TMPL inline bool is_atom (const Symbolic& x) { return is_atom (*x); }
TMPL inline bool is_compound (const Symbolic& x) { return is_compound (*x); }
TMPL inline bool is_sum (const Symbolic& x) { return is_sum (*x); }
TMPL inline bool is_product (const Symbolic& x) { return is_product (*x); }
TMPL inline bool is_func (const Symbolic& x, const Symbolic& f, nat i) {
    return is_func (*x, f.rep, i); }
TMPL inline bool is_func (const Symbolic& x, char* s, nat i) {
    return is_func (*x, s, i); }
TMPL inline bool is_func (const Symbolic& x, const string& s, nat i) {
    return is_func (*x, s, i); }

TMPL inline nat N (const Symbolic& x) { return N (*x); }
TMPL inline nat hash (const Symbolic& x) { return hash (*x); }

/******************************************************************************
* Conversions
******************************************************************************/

TMPL inline C as_scalar (const Symbolic& x) {
  return as_scalar<C> (*x); }
TMPL inline string as_string (const Symbolic& x) {
  return as_string (*x); }

TMPL inline Vector as_vector (const Symbolic& x) {
  vector<generic> v= as_vector (*x);
  return *reinterpret_cast<Vector*> (&v); }
TMPL inline vector<generic> as_vector_generic (const Vector& v) {
  return *reinterpret_cast<const vector<generic>*> (&v); }

TMPL inline Table as_sum (const Symbolic& x) {
  table<generic,generic> t= as_sum (*x);
  return *reinterpret_cast<Table*> (&t); }
TMPL inline Table as_product (const Symbolic& x) {
  table<generic,generic> t= as_product (*x);
  return *reinterpret_cast<Table*> (&t); }
TMPL inline table<generic,generic> as_table_generic (const Table& t) {
  return *reinterpret_cast<const table<generic,generic>*> (&t); }

TMPL inline generic as_generic (const Symbolic& x) {
  return *x; }
TMPL inline ostream& operator << (ostream& out, const Symbolic& x) {
  return out << (*x); }

#undef TMPL_DEF
#undef TMPL
#undef Symbolic
#undef Vector
#undef Table
} // namespace mml
#endif // __SYMBOLIC_HPP
